import 'dart:math';

import 'package:clean_android_design/src/settings/clean_android_design_style.dart';
import 'package:clean_android_design/src/utils/item_section_widget.dart';
import 'package:clean_android_design/src/utils/validation/text_validation.dart';
import 'package:flutter/material.dart';
import 'package:provider/provider.dart';

import 'default_field_widget.dart';

class TextFieldWidget<T> extends StatefulWidget implements IListItem {
  final bool isFirst;
  final bool isLast;
  final Text title;

  final void Function(T, String) saveToItem;
  final String Function(T) getFromItem;

  final TextValidation validation;
  final FocusNode focusNode;
  final FocusNode nextFocusNode;
  final void Function(String) onSubmit;
  final TextInputType keyboardType;
  final bool multiLine;
  final bool separationLine;
  final int maxLines;

  final bool bound;
  final bool isEditable;

  const TextFieldWidget({
    Key key,
    this.saveToItem,
    this.title,
    this.getFromItem,
    this.validation,
    this.focusNode,
    this.nextFocusNode,
    this.onSubmit,
    this.keyboardType,
    this.multiLine = false,
    this.separationLine = true,
    this.maxLines = 8,
  })  : assert(getFromItem != null),
        assert(saveToItem != null),
        isEditable = null,
        bound = true,
        isFirst = false,
        isLast = false,
        super(key: key);

  const TextFieldWidget.unbound({
    Key key,
    this.saveToItem,
    this.title,
    this.getFromItem,
    this.validation,
    this.isEditable,
    this.focusNode,
    this.nextFocusNode,
    this.onSubmit,
    this.keyboardType,
    this.multiLine = false,
    this.separationLine = true,
    this.maxLines = 8,
  })  : assert(getFromItem != null),
        assert(!(isEditable ?? true) || saveToItem != null),
        bound = false,
        isFirst = false,
        isLast = false,
        super(key: key);

  const TextFieldWidget._copy({
    Key key,
    this.isFirst,
    this.isLast,
    this.title,
    this.saveToItem,
    this.getFromItem,
    this.validation,
    this.bound,
    this.isEditable,
    this.focusNode,
    this.nextFocusNode,
    this.onSubmit,
    this.keyboardType,
    this.multiLine = false,
    this.separationLine = true,
    this.maxLines,
  }) : super(key: key);

  @override
  _TextFieldWidgetState<T> createState() => _TextFieldWidgetState<T>();

  @override
  IListItem copyWithPosition({bool isFirst, bool isLast}) {
    return TextFieldWidget<T>._copy(
      key: this.key,
      saveToItem: this.saveToItem,
      getFromItem: this.getFromItem,
      title: this.title,
      validation: this.validation,
      bound: this.bound,
      isEditable: this.isEditable,
      focusNode: this.focusNode,
      nextFocusNode: this.nextFocusNode,
      onSubmit: this.onSubmit,
      keyboardType: this.keyboardType,
      multiLine: this.multiLine,
      separationLine: this.separationLine,
      maxLines: this.maxLines,
      isFirst: isFirst ?? this.isFirst,
      isLast: isLast ?? this.isLast,
    );
  }
}

class _TextFieldWidgetState<T> extends State<TextFieldWidget<T>> {
  T item;
  String errorMessage;

  @override
  void initState() {
    super.initState();

    widget.validation?.initialize(
        getItemFunction: () =>
            widget.bound ? Provider.of<T>(context, listen: false) : null,
        getTextFunction: () => widget.getFromItem(item) ?? "",
        onValidate: (errorMessage) {
          if (mounted) setState(() => this.errorMessage = errorMessage);
        });
  }

  @override
  Widget build(BuildContext context) {
    if (widget.bound) item = Provider.of<T>(context);

    if (widget.isEditable == null)
      return Consumer<bool>(
          builder: (context, isEditable, __) =>
              _getDefautFieldWidget(isEditable));
    else
      return _getDefautFieldWidget(widget.isEditable);
  }

  Widget _getDefautFieldWidget(bool isEditable) {
    var text = widget.getFromItem?.call(item) ?? "";
    var controller = TextEditingController()
      ..value = TextEditingValue(
          text: text,
          selection: TextSelection(
              baseOffset: text.length, extentOffset: text.length));
    controller.addListener(() => widget.saveToItem?.call(
        item, controller.text.trim() == "" ? null : controller.text.trim()));

    var inputBorder = !widget.isLast && widget.separationLine
        ? UnderlineInputBorder(
            borderSide: BorderSide(
                color: CleanAndroidDesignDefaultColors.defaultColorLight,
                width: 1))
        : InputBorder.none;
    var focusedBorder = UnderlineInputBorder(
        borderSide: BorderSide(
      color: CleanAndroidDesignDefaultColors.accent,
      width: 2,
    ));

    return DefaultFieldWidget(
      isFirst: widget.isFirst,
      isLast: widget.isLast,
      separationLine: false,
      title: widget.title,
      childPadding:
          const EdgeInsets.only(left: 30, right: 30, bottom: 8, top: 3),
      child: TextField(
        style: TextStyles.textStyleSize(17)
            .copyWith(color: CleanAndroidDesignDefaultColors.black),
        decoration: InputDecoration(
          contentPadding: EdgeInsets.only(top: 2, bottom: 10),
          isDense: true,
          enabledBorder: inputBorder,
          disabledBorder: inputBorder,
          focusedBorder: focusedBorder,
          errorText: errorMessage,
        ),
        controller: controller,
        cursorColor: CleanAndroidDesignDefaultColors.accentLight,
        enabled: isEditable,
        keyboardType: widget.keyboardType,
        maxLines: widget.multiLine ? widget.maxLines : 1,
        minLines: widget.multiLine
            ? min(max(1, controller.text.split("\n").length),
                (widget.maxLines ?? 100))
            : 1,
        textInputAction: widget.multiLine
            ? TextInputAction.newline
            : widget.nextFocusNode == null
                ? TextInputAction.done
                : TextInputAction.next,
        focusNode: widget.focusNode,
        onSubmitted: (value) {
          if (widget.nextFocusNode != null)
            FocusScope.of(context).requestFocus(widget.nextFocusNode);
          widget.onSubmit?.call(value);
        },
      ),
    );
  }
}
